import re

class HTML(str):
    """A HTML string that will not be escaped."""

    __html__ = _repr_html_ = str.__str__

    def __repr__(self):
        return f"HTML({super().__repr__()})"


def escape(text):
    return HTML(str(text).replace("&", "&amp;").replace("<", "&lt;"))

# Inline styles and scripts only escape the specific end tag
esc_style = re.compile("</(style>)", re.IGNORECASE)
esc_script = re.compile("</(script>)", re.IGNORECASE)

def escape_special(tag: re, text):
    return HTML(tag.sub(r'<\\/\1', text))


def attributes(attrs):
    ret = ""
    for k, v in attrs.items():
        k = mangle(k)
        if v is None or v is False:
            continue
        ret += " " + k
        if v is True:
            continue  # Short attribute
        v = str(v)
        if not v.isalnum():
            v = '"' + v.replace("&", "&amp;").replace('"', "&quot;") + '"'
        ret += "=" + v
    return ret


def mangle(name):
    """Mangle Python identifiers into HTML tag/attribute names.

    Underscores are converted into hyphens. Underscore at end is removed."""
    return name.rstrip("_").replace("_", "-")



from .builder import Builder
from .util import HTML


def Document(*title, _urls=None, _viewport=False, **html_attrs) -> Builder:
    """Construct a new document with a DOCTYPE and minimal structure.

    The html tag is added if any attributes are provided for it.
    If a title is provided, meta charset and title element are added in head.

    E.g. Document("Page title", lang="en") produces a valid document, whereas
    Document() produces only a DOCTYPE declaration.

    Stylesheets, scripts and favicon passed in _urls will be linked in.
    
    Meta viewport may be added to disable device scaling (True) or using a
    custom string value for any other setting."""
    doc = Builder("Document Builder")(HTML("<!DOCTYPE html>"))
    if html_attrs:
        doc.html(**html_attrs)
    if title:
        doc.meta(charset="utf-8")  # Always a good idea
        doc.title(*title)
    if _viewport:
        if _viewport is True:
            _viewport = "width=device-width,initial-scale=1"
        doc.meta(name="viewport", content=_viewport)
    for url in _urls or ():
        fn = url.rsplit("/", 1)[-1]
        ext = fn.rsplit(".", 1)[-1]
        args = linkarg.get(fn) or linkarg.get(ext)
        if args:
            doc.link(href=url, **args)
        elif url.endswith(".js"):
            doc.script(None, src=url, defer=True)
        elif url.endswith(".mjs"):
            doc.script(None, src=url, type="module")
        else:
            raise ValueError("Unknown extension in " + fn)
    return doc

# Arguments for link elements by filename/extension
linkarg = {
    "manifest.json": dict(rel="manifest"),
    "css": dict(rel="stylesheet"),
    "png": dict(rel="icon", type="image/png"),
    "svg": dict(rel="icon", type="image/svg+xml"),
    "ico": dict(rel="icon", type="image/x-icon"),
    "webp": dict(rel="icon", type="image/webp"),
    "avif": dict(rel="icon", type="image/avif"),
}



# HTML5 makes many end tags optional or forbidden (void elements)
void = set("area base br col embed hr img input keygen link menuitem meta param source track wbr".split())
optional = set("html head body p colgroup thead tbody tfoot tr th td li dt dd optgroup option".split())
omit_endtag = void | optional
assert not void & optional


# HTML elements that cannot have content (list incomplete, not used so far)
# - We could automatically close these when opened
empty = set("iframe".split())
assert not empty & omit_endtag
